SystemOrganization addCategory: #'OmniST-Core-Misc'!
SystemOrganization addCategory: #'OmniST-Core-Compiler'!
SystemOrganization addCategory: #'OmniST-Core-Reflection'!
SystemOrganization addCategory: #'OmniST-Core-Compiler-Test'!

!RBMethodNode methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/3/2011 22:26'!
sourceText
	^source! !

!ContextPart methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:18'!
canHandleSignal: exception
	"Sent to handler (on:do:) contexts only.  If my exception class (first arg) handles exception then return true, otherwise forward this message to the next handler context.  If none left, return false (see nil>>canHandleSignal:)"

	<omniDoNotTransform>
	^ (((self tempAt: 1) handles: exception) and: [self tempAt: 3])
		or: [self nextHandlerContext canHandleSignal: exception].
! !

!ContextPart methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:18'!
handleSignal: exception
	"Sent to handler (on:do:) contexts only.  If my exception class (first arg) handles exception then execute my handle block (second arg), otherwise forward this message to the next handler context.  If none left, execute exception's defaultAction (see nil>>handleSignal:)."

	<omniDoNotTransform>
	| val |
	(((self tempAt: 1) handles: exception) and: [self tempAt: 3]) ifFalse: [
		^ self nextHandlerContext handleSignal: exception].

	exception privHandlerContext: self contextTag.
	self tempAt: 3 put: false.  "disable self while executing handle block"
	val := [(self tempAt: 2) cull: exception]
		ensure: [self tempAt: 3 put: true].
	self return: val.  "return from self if not otherwise directed in handle block"
! !

!RBOwnedScope methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/9/2011 15:20'!
numOmniTemps
	| parentCount |
	parentCount := owner ifNil:    [0]
	                     ifNotNil: [owner numOmniTemps].
	^ (bindings keys select: [:name | name beginsWith: 'x_omni_']) size + parentCount! !

!RBOwnedScope methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/8/2011 15:19'!
theClass
	^ owner theClass! !

!String methodsFor: '*OmniST-Core-Compiler-overrides' stamp: 'StefanMarr 12/8/2011 21:56'!
numArgs 
	"Answer either the number of arguments that the receiver would take if considered a selector.  Answer -1 if it couldn't be a selector. It is intended mostly for the assistance of spelling correction."

	| firstChar numColons start ix |
	self size = 0 ifTrue: [^ -1].
	self isOmniSelector 
		ifTrue: [ ^ self asNormalSelector numArgs ].
	firstChar := self at: 1.
	(firstChar isLetter or: [ firstChar = $_]) ifTrue:
		["Fast reject if any chars are non-alphanumeric
		NOTE: fast only for Byte things - Broken for Wide"
		self class isBytes
			ifTrue: [(self findSubstring: '~' in: self startingAt: 1 matchTable: Tokenish) > 0 ifTrue: [^ -1]]
			ifFalse: [2 to: self size do: [:i | (self at: i) tokenish ifFalse: [^ -1]]].
		"Fast colon count"
		numColons := 0.  start := 1.
		[(ix := self indexOf: $: startingAt: start) > 0]
			whileTrue:
				[ix = start ifTrue: [^-1].
				numColons := numColons + 1.
				start := ix + 1].
		numColons = 0 ifTrue: [^ 0].
		self last = $:
			ifTrue: [^ numColons]
			ifFalse: [^ -1]].
	firstChar isSpecial ifTrue:
		[self size = 1 ifTrue: [^ 1].
		2 to: self size do: [:i | (self at: i) isSpecial ifFalse: [^ -1]].
		^ 1].
	^ -1.! !

!String methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 18:48'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	
	<primitive: 105>
	<omniReplacement: #omniReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!MethodNode methodsFor: '*OmniST-Core-Compiler code generation' stamp: 'StefanMarr 11/28/2011 14:36'!
generate: trailer ifQuick: methodBlock
	| v |
	(primitive = 0 and: [arguments size = 0 and: [block isQuick]])
		ifFalse: [^ self].	
	v := block code.
	v < 0
		ifTrue: [^ self].
	v = LdSelf
		ifTrue: [^ self].
	(v between: LdTrue and: LdMinus1 + 3)
		ifTrue: [^ methodBlock value: (CompiledMethod toReturnConstant: v - LdSelf trailerBytes: trailer)].
	v < ((CodeBases at: LdInstType) + (CodeLimits at: LdInstType))
		ifTrue: [^ self].
	^ self.
	! !

!WriteStream methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:33'!
nextPut: anObject
	"Primitive. Insert the argument at the next position in the Stream
	represented by the receiver. Fail if the collection of this stream is not an
	Array or a String. Fail if the stream is positioned at its end, or if the
	position is out of bounds in the collection. Fail if the argument is not
	of the right type for the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 66>
	<omniReplacement: #atomicNextPut:>
	((collection class == ByteString) and: [
		anObject isCharacter and:[anObject isOctetCharacter not]]) ifTrue: [
			collection := (WideString from: collection).
			^self nextPut: anObject.
	].
	position >= writeLimit
		ifTrue: [^ self pastEndPut: anObject]
		ifFalse: 
			[position := position + 1.
			^collection at: position put: anObject]! !

!RBSequenceNode methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/9/2011 15:30'!
addNewOmniTemp: idx
	| allOmniTemps newTemp |
	"allOmniTemps := self allTemporaryVariables select: [:var | var beginsWith: 'x_omni_'].
	newTemp := 'x_omni_', (allOmniTemps size + 1) asString, '_'."
	newTemp := 'x_omni_', idx asString, '_'.
	^ self addTemporaryNamed: newTemp.! !

!Symbol methodsFor: '*OmniST-Core-Compiler-override' stamp: 'StefanMarr 12/11/2011 16:49'!
precedence
	"Answer the receiver's precedence, assuming it is a valid Smalltalk
	message selector or 0 otherwise.  The numbers are 1 for unary,
	2 for binary and 3 for keyword selectors."

	self isOmniSelector 
		ifTrue: [ ^ self asNormalSelector precedence ].
	self size = 0 ifTrue: [^ 0].
	self first isLetter ifFalse: [^ 2].
	self last = $: ifTrue: [^ 3].
	^ 1! !

RBCompiler subclass: #OstCompiler
	instanceVariableNames: 'doFullTransformation'
	classVariableNames: 'OldCompiler'
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler'!

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/7/2011 22:11'!
disableCompiler
	RBCompilerDispatcher current: OldCompiler.! !

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/7/2011 22:11'!
enableCompiler
	OldCompiler := RBCompilerDispatcher current.
	RBCompilerDispatcher current: self.! !

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/8/2011 14:54'!
flushAllOmniMethods
	Smalltalk allClassesAndTraits 
		do: [ :ea | ea flushOmniMethods. ea class flushOmniMethods ]
		displayingProgress: 'Flushing Omnified Methods'.! !

!OstCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/9/2011 16:41'!
initialize
	self enableCompiler.
	self registerChangeListener.
	
	(RBMessageNode hasInstVarNamed: #isOmniNode) ifFalse: [
		RBMessageNode addInstVarNamed: #isOmniNode.
	].! !

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/11/2011 18:30'!
installOmniVersionOf: aCompiledMethod
	"Answer the atomic counterpart of aCompiledMethod.
	Methods with the pragma #omniDoNotTransform are considered to be already omnified.
	Methods with the pragma #omni: will use the annotation code for the omnified counterpart."
	<omniDoNotTransform>
	| omniCM omniName methDict |

	"Is already atomic do nothing"
	aCompiledMethod selector isOmniSelector ifTrue: [
		^ aCompiledMethod ].

	omniName := aCompiledMethod selector asOmniSelector.
	methDict := aCompiledMethod methodClass methodDict.
	
	Transcript cr; show: 'Compiling omnified version of ', omniName, ' in ', aCompiledMethod methodClass name.
	
	"These pragmas are specified in OstTransformer>>isOmniTransformationPragma:"
	aCompiledMethod pragmas do: [ :each |
		each keyword = #omniDoNotTransform
			ifTrue: [ ^ methDict at: omniName put: aCompiledMethod ].
			
		each keyword = #omniReplacement:
			ifTrue: [
				omniCM := self installOmniVersionOf: 
								(aCompiledMethod methodClass lookupSelector: each arguments first).
				^ methDict at: omniName put: omniCM.
			].
	].

	omniCM := self recompileOmnified: aCompiledMethod.
	^ methDict at: omniName put: omniCM! !

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/8/2011 14:20'!
omniMethod: aMessage missing: aReceiver in: aClass
	"This method is called whenever a missing atomic method is called. Installs a whole hierarchy of atomic methods."
	<omniDoNotTransform>
	
	| selector addedAtomicMethod curClass |
	selector := aMessage selector asNormalSelector.
	ProtoObject withAllSubclassesDo: [ :class |
		(class methodDictionary includesKey: selector) ifTrue: [
			self installOmniVersionOf: (class methodDictionary at: selector).] ].
! !

!OstCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/8/2011 14:39'!
onMethodModified: anObject
	<omniDoNotTransform>
	| changedMethod |
	
	"Ignore the already omnified methods"
	anObject itemSelector isOmniSelector
		ifTrue: [^self].
	
	"Also blacklist a few things"
	((anObject itemClass = Debugger)
	or: [anObject itemClass = Decompiler])
		ifTrue: [^ self].

	"For the rest we do the atomic stuff"	
	changedMethod := anObject itemClass methodDict at: anObject itemSelector.
	self installOmniVersionOf: changedMethod.! !

!OstCompiler class methodsFor: 'events' stamp: 'StefanMarr 12/11/2011 16:50'!
onMethodRemoved: anObject
	<omniDoNotTransform>
	
	anObject itemClass methodDict 
		removeKey: anObject itemSelector 
			asOmniSelector
		ifAbsent: nil! !

!OstCompiler class methodsFor: 'compiling' stamp: 'StefanMarr 12/7/2011 22:24'!
recompile: aCompiledMethod
	"Recompile the given method normally"
	<omniDoNotTransform>
	
	| compiler |
	compiler := self new.
	^ compiler recompile: aCompiledMethod.! !

!OstCompiler class methodsFor: 'compiling' stamp: 'StefanMarr 12/7/2011 22:24'!
recompileOmnified: aCompiledMethod
	"Recompile the given method normally"
	<omniDoNotTransform>
	
	| compiler |
	compiler := self new.
	^ compiler recompileOmnified: aCompiledMethod.! !

!OstCompiler class methodsFor: 'compiler support' stamp: 'StefanMarr 12/8/2011 15:33'!
recompilePackageWithOstCompiler
	<omniDoNotTransform>
	
	self enableCompiler.
	self flushAllOmniMethods.

	(PackageInfo named: 'OmniST') classes
		do: [:c | c compileAll]
		displayingProgress: 'Recompiling OmniST with OstCOmpiler support enabled'.! !

!OstCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/8/2011 14:40'!
registerChangeListener

	SystemChangeNotifier uniqueInstance
		"Currently I only do lazy compilation on first use for new stuff"
		"notify: self ofSystemChangesOfItem: #method change: #Added using: #onMethodAdded:;"
		
		"For existing stuff, however, we need to ensure consistency"
		notify: self ofSystemChangesOfItem: #method change: #Modified using: #onMethodModified:;
		notify: self ofSystemChangesOfItem: #method change: #Removed using: #onMethodRemoved:.! !

!OstCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/9/2011 16:42'!
unload
	self flushAllOmniMethods.
	
	RBMessageNode removeInstVarNamed: #isOmniNode.! !

!OstCompiler class methodsFor: 'class initialization' stamp: 'StefanMarr 12/7/2011 22:11'!
unregisterChangeListener

	SystemChangeNotifier uniqueInstance
		noMoreNotificationsFor: self.! !

!OstCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 22:11'!
initialize
	super initialize.
	doFullTransformation := false.! !

!OstCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 22:26'!
parse: methodBoolean
	| methodNode transformer |
	methodNode := super parse: methodBoolean.
	
	methodBoolean ifTrue: [
		transformer := OstTransformer new.
		transformer fullTransformation: doFullTransformation.
		transformer applyTransformations: methodNode inScopeOf: class.
	].
	^ methodNode! !

!OstCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 22:14'!
recompile: aCompiledMethod
	"Recompile the given method normally, i.e., with out full STM transformation"

	^ self recompile: aCompiledMethod omnified: false.! !

!OstCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 16:50'!
recompile: aCompiledMethod omnified: aBool
	"Recompile the given method"

	| methodSource newSelector  newMethodNode |
	
	doFullTransformation := aBool.
	
	methodSource := aCompiledMethod getSource.
	newSelector  := aCompiledMethod selector asOmniSelector.
		
	newMethodNode := self
		compile: methodSource in: aCompiledMethod methodClass notifying: self
		ifFail: [^ self problem: 'syntax error' sel: newSelector in: aCompiledMethod methodClass].

	^ newMethodNode generate.! !

!OstCompiler methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 22:14'!
recompileOmnified: aCompiledMethod
	"Recompile the given method with the STM enabled"

	^ self recompile: aCompiledMethod omnified: true.! !

!Process methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 16:29'!
omniDomain
	^ omniDomain! !

!Process methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 16:29'!
omniDomain: aDomain
	omniDomain := aDomain! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/8/2011 00:02'!
beOmniNode
	isOmniNode := true.! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'JorgeRessia 11/25/2009 13:33'!
isInlineAndOr
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(self selectorParts allSatisfy: [ :each | #(and: or:) includes: each value ]) 
		ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each isBlock ])  ifFalse: [^ false].
	(self arguments allSatisfy: [ :each | each arguments isEmpty ])
			ifFalse: [
				self notify: 'and: (or:) takes zero-arg block'.
				^ false ].
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineCase

	self isCascaded ifTrue: [^ false].
	(#(caseOf: caseOf:otherwise:) includes: self selector) ifFalse: [^ false].
	self arguments size = 2 ifTrue: [  "otherwise block"
		self arguments last isBlock ifFalse: [^ false]].
	self arguments first isArray ifFalse: [^ false].
	self arguments first statements do: [:assoc |
		(assoc isMessage and: [assoc selector == #->]) ifFalse: [^ false].
		assoc receiver isBlock ifFalse: [^ false].
		assoc receiver arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'. ^ false].
		assoc arguments first isBlock ifFalse: [^ false].
		assoc arguments first arguments isEmpty ifFalse: [self notify: 'caseOf: takes zero-arg blocks'].
	].
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineIf

	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	self arguments do: [:block |
		block arguments isEmpty ifFalse: [
			self notify: 'ifTrue:ifFalse: takes zero-arg blocks'.
			^ false
		]
	].
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'MarcusDenker 8/21/2010 13:40'!
isInlineIfNil

	| assertNone assertOneOrNone |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(ifNil: ifNotNil: ifNil:ifNotNil: ifNotNil:ifNil:) includes: self selector) ifFalse: [^ false].
	self arguments do: [:node | node isBlock ifFalse: [^ false]].
	assertNone := [:block | 
		block arguments isEmpty ifFalse: [self notify: 'ifNil: takes zero-arg block'. ^ false]
	].
	assertOneOrNone := [:block | 
		block arguments size > 1 ifTrue: [self notify: 'ifNotNil: takes zero- or one-arg block'. ^ false]
	].
	self selector caseOf: {
		[#ifNil:] 		 	-> [assertNone value: self arguments first].
		[#ifNil:ifNotNil:] 	-> [assertNone value: self arguments first. 
							   assertOneOrNone value: self arguments last].
		[#ifNotNil:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNilDo:] 		  	-> [assertOneOrNone value: self arguments first].
		[#ifNotNil:ifNil:] 	-> [assertOneOrNone value: self arguments first. 
							   assertNone value: self arguments last]
	}.
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'MarcusDenker 9/3/2010 15:03'!
isInlineToDo

	| block step |
	self receiver isBlock ifTrue: [^ false].
	self isCascaded ifTrue: [^ false].
	(#(to:do: to:by:do:) includes: self selector) ifFalse: [^ false].
	self arguments first isBlock ifTrue: [^ false].
	block := self arguments last.
	block isBlock ifFalse: [^ false].
	block arguments size = 1 ifFalse: [
		self notify: 'to:do: block must take one arg'. ^ false].
	(RBParseTreeSearcher new
		matches: block arguments first name , ' := `@object' do: [:n :a | true];
		executeTree: block initialAnswer: false) ifTrue: [^ false].
	self arguments size = 3 "to:by:do:" ifTrue: [
		step := self arguments second.
		step isLiteral ifFalse: [^ false].
		step value = 0 ifTrue: [self notify: 'by: step can''t be zero'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'jorgeRessia 11/20/2009 16:40'!
isInlineWhile

	self isCascaded ifTrue: [^ false].
	(#(whileFalse: whileTrue: whileFalse whileTrue) includes: self selector) ifFalse: [^ false].
	self receiver isBlock ifFalse: [^ false].
	self receiver arguments isEmpty ifFalse: [self notify: 'while receiver block must have no arguments'. ^ false].
	self arguments isEmpty ifFalse: [
		self arguments first isBlock ifFalse: [^ false].
		self arguments first arguments isEmpty ifFalse: [self notify: 'while takes a zero-arg block as its argument'. ^ false].
	].
	^ true! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler-inline' stamp: 'StefanMarr 12/3/2011 14:22'!
isInlined
	^ self isInlineAndOr or:
	 [self isInlineCase  or:
	 [self isInlineIf    or:
	 [self isInlineIfNil or:
	 [self isInlineToDo  or:
	  self isInlineWhile    ]]]]! !

!RBMessageNode methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/8/2011 00:06'!
isOmniNode
	^ isOmniNode = true "with this, we do not need to change initialize"! !

!MethodContext methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:35'!
tempAt: index 
	"Answer the value of the temporary variable whose index is the 
	 argument, index.  Primitive. Assumes receiver is indexable. Answer the
	 value of an indexable element in the receiver. Fail if the argument index
	 is not an Integer or is out of bounds. Essential. See Object documentation
	 whatIsAPrimitive.  Override the default at: primitive to give latitude to the
	 VM in context management."
	
	<omniDoNotTransform>
	<primitive: 210>
	^self at: index! !

!MethodContext methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:35'!
tempAt: index put: value 
	"Store the argument, value, as the temporary variable whose index is the 
	 argument, index.  Primitive. Assumes receiver is indexable. Answer the
	 value of an indexable element in the receiver. Fail if the argument index
	 is not an Integer or is out of bounds. Essential. See Object documentation
	 whatIsAPrimitive.  Override the default at:put: primitive to give latitude to
	 the VM in context management."

	<omniDoNotTransform>
	<primitive: 211>
	^self at: index put: value! !

RBSemanticAnnotator subclass: #OstTransformer
	instanceVariableNames: 'doFullTransformation doLocalTransformation numOfOmniTemps'
	classVariableNames: 'RequestExecutionSelectors'
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler'!

!OstTransformer commentStamp: '<historical>' prior: 0!
OmniTransformer iterates over the AST of a method to transform it.
The goal is to delegate all read, write, and message-send operations via a domain object.!

!OstTransformer class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 01:04'!
generateRequestExecutionSelectors
	| sel |
	sel := #requestExecutionOf:on:.
	RequestExecutionSelectors := Array new: 10.
	1 to: 10 do: [:i |
		RequestExecutionSelectors at: i put: sel.
		sel := (sel, #with:) asSymbol.
	].! !

!OstTransformer class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 00:58'!
initialize
	self generateRequestExecutionSelectors.! !

!OstTransformer methodsFor: 'visitor-double dispatching' stamp: 'StefanMarr 12/7/2011 20:49'!
acceptAssignmentNode: anAssignmentNode
	super acceptAssignmentNode: anAssignmentNode.
	
	self doTransformation ifFalse: [^self].
	
	(anAssignmentNode variable isInstanceBinding or:
	[anAssignmentNode variable isLiteralBinding "or: REMOTE NOT YET SUPPORTED
	[anAssignmentNode variable isRemote]"]) ifTrue: [
		| replacement |
		replacement := self createOmniAssignment: anAssignmentNode.
		anAssignmentNode parent replaceNode: anAssignmentNode withNode: replacement.
	]! !

!OstTransformer methodsFor: 'visitor-double dispatching' stamp: 'StefanMarr 12/8/2011 15:39'!
acceptMessageNode: aMessageNode
	| doLocalTransformationOld replacement |
	doLocalTransformationOld := doLocalTransformation.
	
	doLocalTransformation := doLocalTransformation or: [self isOmniEntryMessage: aMessageNode selector].

	(self doTransformation and: [aMessageNode isInlined not])
		ifTrue:  [self  transformMessageNode: aMessageNode.]
		ifFalse: [super acceptMessageNode:    aMessageNode.].

	doLocalTransformation := doLocalTransformationOld.! !

!OstTransformer methodsFor: 'visitor-dispatching' stamp: 'StefanMarr 12/7/2011 20:55'!
acceptMethodNode: aMethodNode
	super acceptMethodNode: aMethodNode.
	doFullTransformation ifTrue: [
		aMethodNode selector: (aMethodNode selector asOmniSelector)
	].! !

!OstTransformer methodsFor: 'visitor-dispatching' stamp: 'StefanMarr 12/8/2011 00:56'!
acceptVariableNode: aVariableNode
	| replacement |
	self doTransformation ifFalse: [^self].
	
	super acceptVariableNode: aVariableNode.
	
	"Not every variable node we are going to encounter is a read.
	 Make sure it is not the variable node of a parent assignment node.
	 Furthermore, we do not need to take care of temp assignments.
	 Only globals, instances, and perhaps remote onces of a closure."
	(aVariableNode isRead
	 and: [aVariableNode isLiteralBinding   or: 
		   [aVariableNode isInstanceBinding "or: REMOTES NOT SUPPORTED at the moment
			[aVariableNode isRemote]"]])
		ifFalse: [^self].
	
	"Now we are sure that we have to transform a variable read".
	replacement := self createOmniReadNode: aVariableNode.
	aVariableNode parent replaceNode: aVariableNode withNode: replacement. 
! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/7/2011 20:57'!
applyTransformations: methodNode inScopeOf: aClass
	"We have two cases:
	   Either it is a normal method, and we do only the omni blocks,
	   or it is a fully omni method and we transform everything.
	
	 doFullTransformation indicates that we do not only local adaptations.
	"
	| classScope |
	(self pragmasForbidTransformation: methodNode)
		ifTrue: [^self].
	
	classScope :=  (RBVariableScope 
		owner: (RBLiteralScope 	owner: RBRootScope new class: aClass)
		class: aClass).
	
	self start: methodNode scope: classScope.! !

!OstTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 15:48'!
createObtainCurrentOmniDomainNode
	| msgSendNode  |
	msgSendNode := RBMessageNode receiver: (RBLiteralValueNode value: OstDomain)
									   selector: #currentDomain.
	msgSendNode beOmniNode.
	^ msgSendNode.! !

!OstTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 11:42'!
createObtainOmniDomainNode: receiver
	| msgSendNode |
	msgSendNode := RBMessageNode receiver: receiver
									   selector: #omniDomain.
	msgSendNode beOmniNode.
	^ msgSendNode.! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/8/2011 00:11'!
createOmniAssignment: anAssignmentNode
	| var |
	var := anAssignmentNode variable.
	var isInstanceBinding ifTrue: [
		^ self createWriteMessageSendFor: var with: anAssignmentNode value.
	]
	ifFalse: [
		| globalAssoc sendArgs |
		var isLiteralBinding ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		
		globalAssoc := RBLiteralValueNode value: var variableBinding binding.
		
		sendArgs := OrderedCollection new: 1.
		sendArgs add: anAssignmentNode value.
		^ RBMessageNode receiver: globalAssoc selector: #atomicValue: arguments: sendArgs.
	].! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/8/2011 15:20'!
createOmniDomainRefNode
	"Return a node that will resolve at runtime to the omni domain object,
	 or in case the class does not have a slot called omniDomain, resolve to
	 the nil object"
	| omniIdx omniRef |
	omniIdx := OstDomain domainIndexFromClass: scope theClass.
	omniIdx = 0
		ifTrue: [
			"The class does not have a omniDomain slot, so we just return a special literal"
			omniRef := RBLiteralValueNode value: nil.
		]
		ifFalse:[
			"This class got a omniDomain slot"
			omniRef := RBVariableNode named: #omniDomain.
			omniRef variableBinding: (scope lookup: #omniDomain).
		].
	
	^ omniRef! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/9/2011 15:27'!
createOmniMessageNode: aMessageNode
	| omniSel assignToTempNode surroundingSeq resultTemp resultTempBinding |

	"This one is a more tricky operation.
	 We need to introduce to introduce a temporay to the surrounding sequence.
	 The temporary is the result of the receiver.
	 From which we need to obtain the omniDomain (by sending #omniDomain)
	 and which is passed on as the on: parameter for the actual message send.
	 
	 REM: at the moment, I think we will not need to take special care of arguments that
	      are the result of message sends, their receivers will be treated equally, but
	      the actual message send result can still be handled without temporary.
	 Q: is one temp for all omni sends enough? probably not, go with one per send for now"
	
	omniSel := aMessageNode selector asOmniSelector.
	
	"We only need to go through the trouble if the receiver is not an imediate value"
	aMessageNode receiver isImmediateNode ifTrue:[
		" replace receiver with message send to receiver with #omniDomain"
		" then assemble normal message send or reqExecOn..."
		^ self createRequestExecutionNodeOn: aMessageNode receiver
											 with: omniSel
											  and: aMessageNode arguments.
	].

	"Now lets create a temporary for the result of the receiver"
	surroundingSeq := aMessageNode statementNode parent.	
	resultTemp := surroundingSeq addNewOmniTemp: self nextOmniTempIdx.
	
	resultTempBinding := scope add: (RBTemporaryBinding node: resultTemp).

	assignToTempNode := RBAssignmentNode variable: resultTemp value: aMessageNode receiver.
	
	surroundingSeq addNode: assignToTempNode before: aMessageNode statementNode.
	
	^ self createRequestExecutionNodeOn: resultTemp
										 with: omniSel
										  and: aMessageNode arguments.! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/9/2011 15:50'!
createOmniReadNode: aVariableNode
	| selfRef omniDomainRef instIdx sendArgs |
	
	aVariableNode isInstanceBinding ifTrue: [
		^ self createReadMessageSendFor: aVariableNode.
	]
	ifFalse: [
		aVariableNode isLiteralBinding
			ifFalse: [self halt: 'I guess we still need to implement something, e.g., remote vars.'].
		^ self createReadGlobalMessageSendFor: aVariableNode.
	].
	! !

!OstTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 15:50'!
createReadGlobalMessageSendFor: aVariableNode
	| globalAssoc args sendNode |

	globalAssoc := RBLiteralValueNode value: aVariableNode variableBinding binding.
		
	args := OrderedCollection new: 1.
	args add: globalAssoc.
		
	sendNode := RBMessageNode receiver: self createObtainCurrentOmniDomainNode
		                          selector: #readGlobal:
		                          arguments: args.
	sendNode beOmniNode.
	^ sendNode! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/8/2011 01:21'!
createReadMessageSendFor: aVariableNode
	| selfRef omniDomainRef instIdx sendArgs readNode |
	selfRef := self createSelfRefNode.
	omniDomainRef := self createOmniDomainRefNode.
	
	instIdx  := RBLiteralValueNode value: (aVariableNode variableBinding index).
	sendArgs := OrderedCollection new: 2.
	sendArgs add: instIdx.
	sendArgs add: selfRef.
	readNode := RBMessageNode receiver: omniDomainRef selector: #readField:of: arguments: sendArgs.
	readNode beOmniNode.
	^ readNode
! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/9/2011 11:45'!
createRequestExecutionNodeOn: receiver with: selector and: arguments
	| msgSendNode reqExecSelector msgArguments numArgs omniDomain|
	
	omniDomain := self createObtainOmniDomainNode: receiver.
	
	" #requestExecutionOf:on:with:... "
	numArgs := selector numArgs.
	reqExecSelector := self requestExecutionSelectorFor: numArgs.
	
	msgArguments := OrderedCollection new: numArgs + 2.
	msgArguments add: (RBLiteralValueNode value: selector).
	msgArguments add: receiver.
	msgArguments addAll: arguments.
	
	msgSendNode := RBMessageNode receiver: omniDomain
								   selector: reqExecSelector
								   arguments: msgArguments.
	msgSendNode beOmniNode.
	^ msgSendNode! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/8/2011 01:11'!
createSelfRefNode
	| selfRef |
	selfRef := RBVariableNode named: 'self'.
	selfRef variableBinding: (scope lookup: 'self').
	^ selfRef! !

!OstTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 21:57'!
createWriteMessageSendFor: aVarNode with: aValue
		| selfRef omniDomainRef instIdx sendArgs sendNode |
		selfRef := self createSelfRefNode.
		omniDomainRef := self createOmniDomainRefNode.
		
		instIdx := RBLiteralValueNode value: (aVarNode variableBinding index).
		
		sendArgs := OrderedCollection new: 3.
		sendArgs add: aValue.
		sendArgs add: instIdx.
		sendArgs add: selfRef.
		
		sendNode := RBMessageNode receiver: omniDomainRef
									   selector: #write:toField:of:
									   arguments: sendArgs.
		sendNode beOmniNode.
		^ sendNode
! !

!OstTransformer methodsFor: 'testing' stamp: 'StefanMarr 12/7/2011 20:35'!
doTransformation
	^ doFullTransformation or: [doLocalTransformation]! !

!OstTransformer methodsFor: 'accessing' stamp: 'StefanMarr 12/7/2011 20:35'!
fullTransformation: aBool
	doFullTransformation := aBool! !

!OstTransformer methodsFor: 'initialize-release' stamp: 'StefanMarr 12/9/2011 15:29'!
initialize
	super initialize.
	doFullTransformation  := false.
	doLocalTransformation := false.
	numOfOmniTemps := 0.
! !

!OstTransformer methodsFor: 'visitor-double dispatching' stamp: 'StefanMarr 12/9/2011 17:01'!
isOmniEntryMessage: aSelector
	^ aSelector = #omni or: [aSelector = #omni: or: [aSelector = #atomic]].! !

!OstTransformer methodsFor: 'testing' stamp: 'StefanMarr 12/8/2011 15:44'!
isOmniTransformationPragma: aSelector
	^ aSelector = #omniDoNotTransform or: 
	 [aSelector = #omniReplacement:]! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/9/2011 15:29'!
nextOmniTempIdx
	"Creates a sequence of unique id numbers for a method transformation to name the temporary variables"
	numOfOmniTemps := numOfOmniTemps + 1.
	^ numOfOmniTemps! !

!OstTransformer methodsFor: 'testing' stamp: 'StefanMarr 12/8/2011 15:42'!
pragmasForbidTransformation: methodNode
	"Some of our pragmas restrict the boundaries of the Omni execution layer.
	 They enable to break out into the implementation level."
	
	methodNode pragmas do: [:each |
		(self isOmniTransformationPragma: each selector)
			ifTrue: [^true].
	].
	^false.! !

!OstTransformer methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 15:43'!
requestExecutionSelectorFor: nArgs
	(nArgs >= 0 and: [nArgs < 10])
		ifTrue: [
			^ RequestExecutionSelectors at: nArgs + 1.
		]
		ifFalse: [
			self halt: 'This is currently not implemented!! ',
			           'Need either to extend this table/array, ',
			           'or make a distinction to use the other primitive (84) which accepts an array.'.
		].
	^ nil! !

!OstTransformer methodsFor: 'transformation' stamp: 'StefanMarr 12/8/2011 15:44'!
transformMessageNode: aMessageNode
	super acceptMessageNode: aMessageNode.

	((self isOmniEntryMessage: aMessageNode selector) or: [
	 aMessageNode isInlined or: [
	 aMessageNode isOmniNode]]) 
		ifFalse: [
			| replacement |
			replacement := self createOmniMessageNode: aMessageNode.
			aMessageNode parent replaceNode: aMessageNode withNode: replacement.
		].! !

!Behavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 18:10'!
basicNew
	"Primitive. Answer an instance of the receiver (which is a class) with no 
	indexable variables. Fail if the class is indexable. Essential. See Object 
	documentation whatIsAPrimitive."

	<primitive: 70>
	<omniReplacement: #omniBasicNew>
	self isVariable ifTrue: [ ^ self basicNew: 0 ].
	"space must be low"
	OutOfMemory signal.
	^ self basicNew  "retry if user proceeds"! !

!Behavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 18:11'!
basicNew: sizeRequested 
	"Primitive. Answer an instance of this class with the number
	of indexable variables specified by the argument, sizeRequested.
	Fail if this class is not indexable or if the argument is not a
	positive Integer, or if there is not enough memory available. 
	Essential. See Object documentation whatIsAPrimitive."

	<primitive: 71>
	<omniReplacement: #omniBasicNew:>
	self isVariable ifFalse:
		[self error: self printString, ' cannot have variable sized instances'].
	(sizeRequested isInteger and: [sizeRequested >= 0]) ifTrue:
		["arg okay; space must be low."
		OutOfMemory signal.
		^ self basicNew: sizeRequested  "retry if user proceeds"].
	self primitiveFailed! !

!Behavior methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/8/2011 14:55'!
flushOmniMethods
	| copy |
	self methodDictionary keysDo: [ :each |
		each isOmniSelector ifTrue: [
			copy := copy ifNil: [ self methodDictionary copy ].
			copy removeDangerouslyKey: each ifAbsent: [] ] ].
	copy ifNotNil: [ self methodDictionary become: copy ]! !

!Behavior methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/11/2011 18:10'!
omniBasicNew
	| obj |
	<omniDoNotTransform>
	obj := self basicNew.
	OstDomain initializeDomainForNewObject: obj of: self.
	^ obj! !

!Behavior methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/11/2011 18:11'!
omniBasicNew: sizeRequested 
	"This is the omni-fied implementation.
	 The original basicNew: is now rawBasicNew:."
	<omniDoNotTransform>
	
	| obj |
	obj := self basicNew: sizeRequested.
	OstDomain initializeDomainForNewObject: obj of: self.
	^ obj! !

!Behavior methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/9/2011 17:08'!
omniMethods
	^ Array streamContents: [ :stream |
		self methodDictionary keysAndValuesDo: [ :key :value |
			key isOmniSelector 
				ifTrue: [ stream nextPut: value ] ] ]! !

!Behavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/9/2011 17:05'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys reject: [ :each | each isOmniSelector ]! !

!Behavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/9/2011 17:05'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isOmniSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!Behavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/9/2011 17:05'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: [ :selector |
		selector isOmniSelector 
			ifFalse: [ selectorBlock value: selector ] ]! !

!RBVariableNode methodsFor: '*OmniST-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:20'!
isInstanceBinding
	^ self variableBinding isInstanceBinding! !

!RBVariableNode methodsFor: '*OmniST-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:27'!
isLiteralBinding
	^ self variableBinding isLiteralBinding! !

!RBVariableNode methodsFor: '*OmniST-Core-Compiler-ast-semantic-accessing' stamp: 'StefanMarr 12/4/2011 15:49'!
variableBinding: aBinding
	^ self propertyAt: #variableBinding put: aBinding! !

!ProcessorScheduler methodsFor: '*OmniST-Core-Roar-Compatibility' stamp: 'StefanMarr 11/6/2011 14:46'!
thisProcess
	"Added for RVM compatibility already.
	 Just to make sure that I can switch easily."
	^ self activeProcess! !

!TPureBehavior methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/8/2011 14:58'!
flushOmniMethods
	| copy |
	self methodDictionary keysDo: [ :each |
		each isOmniSelector ifTrue: [
			copy := copy ifNil: [ self methodDictionary copy ].
			copy removeDangerouslyKey: each ifAbsent: [] ] ].
	copy ifNotNil: [ self methodDictionary become: copy ]! !

!TPureBehavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:44'!
selectors
	"Answer a Set of all the message selectors specified in the receiver's 
	method dictionary."

	^ self methodDict keys reject: [ :each | each isOmniSelector ]! !

!TPureBehavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:44'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isOmniSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

!TPureBehavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:44'!
selectorsDo: selectorBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysDo: [ :selector |
		selector isOmniSelector 
			ifFalse: [ selectorBlock value: selector ] ]! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 21:53'!
evaluateInContext: aBlock
	<omniDoNotTransform>
	^ OstDomain new evaluateInContext: aBlock! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 21:53'!
readField: idx of: obj
	<omniDoNotTransform>
	^ obj instVarAt: idx! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 21:53'!
requestExecutionOf: aSelector on: anObject
	<omniDoNotTransform>
	^ anObject perform: aSelector! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 21:53'!
requestExecutionOf: aSelector on: anObject with: p1
	<omniDoNotTransform>
	^ anObject perform: aSelector with: p1! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 22:00'!
requestExecutionOf: aSelector on: anObject with: p1 with: p2
	<omniDoNotTransform>
	^ anObject perform: aSelector with: p1 with: p2! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 22:00'!
requestExecutionOf: aSelector on: anObject with: p1 with: p2 with: p3
	<omniDoNotTransform>
	^ anObject perform: aSelector with: p1 with: p2 with: p3! !

!UndefinedObject methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/9/2011 15:58'!
requestExecutionOf: aSelector on: anObject with: p1 with: p2 with: p3 with: p4
	<omniDoNotTransform>
	^ anObject perform: aSelector with: p1 with: p2 with: p3 with: p4! !

!BlockContext methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:15'!
tempAt: index 
	"Refer to the comment in ContextPart|tempAt:."

	<omniDoNotTransform>
	^ home at: index! !

!BlockContext methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:15'!
tempAt: index put: value 
	"Refer to the comment in ContextPart|tempAt:put:."

	<omniDoNotTransform>
	^ home at: index put: value! !

!ReadStream methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:33'!
next
	"Primitive. Answer the next object in the Stream represented by the
	receiver. Fail if the collection of this stream is not an Array or a String.
	Fail if the stream is positioned at its end, or if the position is out of
	bounds in the collection. Optional. See Object documentation
	whatIsAPrimitive."

	<primitive: 65>
	<omniReplacement: #atomicNext>
	position >= readLimit
		ifTrue: [^nil]
		ifFalse: [^collection at: (position := position + 1)]! !

!BlockClosure methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 16:00'!
omni
	^ OstDomain currentDomain evaluateInContext: self.! !

!BlockClosure methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/8/2011 16:01'!
omni: aDomain
	^ aDomain evaluateInContext: self.! !

!Integer methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:35'!
timesRepeat: aBlock
	"Evaluate the argument, aBlock, the number of times represented by the 
	receiver."

	<omniDoNotTransform>
	| count |
	count := 1.
	[count <= self]
		whileTrue: 
			[aBlock value.
			count := count + 1]! !

!Exception methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:35'!
resumeUnchecked: resumptionValue
	"Return resumptionValue as the value of #signal, unless this was called after an #outer message, then return resumptionValue as the value of #outer."
	<omniDoNotTransform>
	| ctxt |
	outerContext 
		ifNil: [
			signalContext return: resumptionValue ] 
		ifNotNil: [
			ctxt := outerContext.
			outerContext := ctxt tempAt: 1. "prevOuterContext in #outer"
			ctxt return: resumptionValue ]
! !

!Array methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/9/2011 17:03'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 105>
	<omniReplacement: #omniReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

TestCase subclass: #OstCompilerTest
	instanceVariableNames: 'foo'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler-Test'!

!OstCompilerTest methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/28/2011 13:17'!
basicReadOn: o
	self assert: 'AB' equals: o aAndB. " This one first since that is usually not optimizied by the compiler"

	self assert: #A equals: o a.
	self assert: #B equals: o b.! !

!OstCompilerTest methodsFor: 'test helpers' stamp: 'StefanMarr 12/8/2011 22:39'!
basicReadOn: o in: aDomain
	" This one first since that is usually not optimizied by the compiler"

	self assert: 'AB' equals: ([ o aAndB ] omni: aDomain). 

	self assert: #A equals: ([ o a ] omni: aDomain).
	self assert: #B equals: ([ o b ] omni: aDomain).! !

!OstCompilerTest methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:58'!
constReadOn: o
	self assert: 1 equals: o return1.
	self assert: nil equals: o returnNil.
	self assert: o equals: o returnSelf.! !

!OstCompilerTest methodsFor: 'test helpers' stamp: 'StefanMarr 12/8/2011 22:39'!
constReadOn: o in: aDomain
	self assert: 1   equals: ([ o return1    ] omni: aDomain).
	self assert: nil equals: ([ o returnNil  ] omni: aDomain).
	self assert: o   equals: ([ o returnSelf ] omni: aDomain).! !

!OstCompilerTest methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:59'!
otherReadsOn: o
	self assert: #A equals: o returnAViaTemp.
	
	self assert: #A equals: o somethingLongAndComplexReturningOriginalA.! !

!OstCompilerTest methodsFor: 'test helpers' stamp: 'StefanMarr 12/9/2011 16:31'!
otherReadsOn: o in: aDomain
	self assert: #A equals: ([ o returnAViaTemp ] omni: aDomain).
	
	self flag: 'TODO: check whether there is a way to get this one working, by reducing the number of created temp vars'.
	[self assert: #A equals: ([ o somethingLongAndComplexReturningOriginalA ] omni: aDomain).].! !

!OstCompilerTest methodsFor: 'running' stamp: 'StefanMarr 11/28/2011 14:20'!
setUp
	[OstCompilerTestSubject compileAll].
	Transcript log: 'Make sure the test target is compiled properly. Using compileAll does not seem to trigger properly the mechanism used for replacing methods by methodwrappers, i.e., there are not notifications.'.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 22:40'!
testAllMethodSemanticsWithoutDomain
	| o |
	o := OstCompilerTestSubject new.
	
	o omniDomain: nil. "Not strictly necessary, but to be explicit"
	
	self basicReadOn:  o in: nil.
	self otherReadsOn: o in: nil.
	self constReadOn:  o in: nil.
	self writesOn:     o in: nil.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/9/2011 14:14'!
testBasicReadWithDomain
	| o domain |
	o := OstCompilerTestSubject new.
	
	domain := OstCompilerTestDomain new.
	o omniDomain: domain.
	
	self basicReadOn: o in: domain.

	self assert: {#requestExecutionOf:on:. #aAndB. o } equals: domain record removeFirst.
	self assert: {#readField:of:. 2. o.} equals: domain record removeFirst.
	self assert: {#readField:of:. 3. o.} equals: domain record removeFirst.
	
	self assert: {#requestExecutionOf:on:. #a. o } equals: domain record removeFirst.
	self assert: {#readField:of:. 2. o.} equals: domain record removeFirst.
	
	self assert: {#requestExecutionOf:on:. #b. o } equals: domain record removeFirst.
	self assert: {#readField:of:. 3. o.} equals: domain record removeFirst.
	
! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 23:46'!
testBasicSend
	| result |
	result := [ 1 bitOr: 2 ] omni.
	self assert: 3 equals: result.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 23:49'!
testChainedAndComplexArgumentSend
	| result |
	result := [ (1 bitOr: 2) bitOr: (8 bitOr: 16) ] omni.
	self assert: 27 equals: result.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 23:47'!
testChainedSend
	| result |
	result := [ (1 bitOr: 2) bitOr: 4 ] omni.
	self assert: 7 equals: result.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 23:49'!
testComplexArgumentSend
	| result |
	result := [ 1 bitOr: (8 bitOr: 16) ] omni.
	self assert: 25 equals: result.! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/8/2011 22:37'!
testConstReadWithDomain
	| o domain |
	o := OstCompilerTestSubject new.
	
	domain := OstCompilerTestDomain new.
	o omniDomain: domain.
	
	self constReadOn: o in: domain.
		
	self assert: {#requestExecutionOf:on:. #return1.    o} equals: domain record removeFirst.
	self assert: {#requestExecutionOf:on:. #returnNil.  o} equals: domain record removeFirst.
	self assert: {#requestExecutionOf:on:. #returnSelf. o} equals: domain record removeFirst.
	! !

!OstCompilerTest methodsFor: 'tests' stamp: 'StefanMarr 12/9/2011 16:34'!
testOtherReadsWithDomain
	| o domain |
	o := OstCompilerTestSubject new.
	
	domain := OstCompilerTestDomain new.
	o omniDomain: domain.
	
	self otherReadsOn: o in: domain.
	
	self assert: {#requestExecutionOf:on:. #returnAViaTemp. o } equals: domain record removeFirst.
	self assert: {#readField:of:. 2. o.} equals: domain record removeFirst.
	
	
	self flag: 'TODO: currently disabled since this test does not fit into the temp boundaries of smalltalk :('.
	[ self assert: {#requestExecutionOf:on:. #somethingLongAndComplexReturningOriginalA. o. }
	     equals: domain record removeFirst.
     self assert: {#readField:of:. 2. o.} equals: domain record removeFirst.].
	! !

!OstCompilerTest methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:59'!
writesOn: o
	self deny: (o a = 42).
	o a: 42.
	self assert: 42 equals: o a.
	
	self deny: (o b = 55).
	o b: 55.
	self assert: 55 equals: o b.
	! !

!OstCompilerTest methodsFor: 'test helpers' stamp: 'StefanMarr 12/8/2011 22:41'!
writesOn: o in: aDomain
	self deny: (o a = 42).
	o a: 42.
	self assert: 42 equals: ([o a] omni: aDomain).
	
	self deny: (o b = 55).
	o b: 55.
	self assert: 55 equals: ([o b] omni: aDomain).
	! !

TestCase subclass: #OstCompilerUtilitiesTest
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler-Test'!

!OstCompilerUtilitiesTest methodsFor: 'testing' stamp: 'StefanMarr 12/11/2011 16:47'!
testAsNormalSelector
	self assert: #'xx_omni_x_string' asNormalSelector == #string.
	self assert: #'xx_omni_x_linesDo:' asNormalSelector == #linesDo:.
	self assert: #'xx_omni_x_indexOf:startingAt:' asNormalSelector == #indexOf:startingAt:.
	self assert: #'xx_omni_x_//' asNormalSelector == #//! !

!OstCompilerUtilitiesTest methodsFor: 'testing' stamp: 'StefanMarr 12/11/2011 16:47'!
testAsOmniSelector
	self assert: #string asOmniSelector == #'xx_omni_x_string'.
	self assert: #linesDo: asOmniSelector == #'xx_omni_x_linesDo:'.
	self assert: #indexOf:startingAt: asOmniSelector == #'xx_omni_x_indexOf:startingAt:'.
	self assert: #// asOmniSelector == #'xx_omni_x_//'! !

!OstCompilerUtilitiesTest methodsFor: 'testing-symbol' stamp: 'StefanMarr 12/11/2011 16:47'!
testIsInfix
	self deny: #string isInfix.
	self deny: #'xx_omni_x_string' isInfix.
	
	self deny: #linesDo: isInfix.
	self deny: #'xx_omni_x_linesDo:' isInfix.
	
	self deny: #indexOf:startingAt: isInfix.
	self deny: #'xx_omni_x_indexOf:startingAt:' isInfix.
	
	self assert: #// isInfix.
	self assert: #'xx_omni_x_//' isInfix! !

!OstCompilerUtilitiesTest methodsFor: 'testing-symbol' stamp: 'StefanMarr 12/11/2011 16:48'!
testIsKeyword
	self deny: #string isKeyword.
	self deny: #'xx_omni_x_string' isKeyword.
	
	self assert: #linesDo: isKeyword.
	self assert: #'xx_omni_x_linesDo:' isKeyword.
	
	self assert: #indexOf:startingAt: isKeyword.
	self assert: #'xx_omni_x_indexOf:startingAt:' isKeyword.
	
	self deny: #// isKeyword.
	self deny: #'xx_omni_x_//' isKeyword! !

!OstCompilerUtilitiesTest methodsFor: 'testing' stamp: 'StefanMarr 12/11/2011 17:18'!
testIsOmniSelector
	self deny: #string isOmniSelector.
	self deny: #linesDo: isOmniSelector.
	self deny: #indexOf:startingAt: isOmniSelector.
	self deny: #// isOmniSelector.

	self assert: #'xx_omni_x_string' isOmniSelector.
	self assert: #'xx_omni_x_linesDo:' isOmniSelector.
	self assert: #'xx_omni_x_indexOf:startingAt:' isOmniSelector.
	self assert: #'xx_omni_x_//:' isOmniSelector! !

!OstCompilerUtilitiesTest methodsFor: 'testing-symbol' stamp: 'StefanMarr 12/11/2011 16:48'!
testIsUnary
	self assert: #string isUnary.
	self assert: #'xx_omni_x_string' isUnary.
	
	self deny: #linesDo: isUnary.
	self deny: #'xx_omni_x_linesDo:' isUnary.
	
	self deny: #indexOf:startingAt: isUnary.
	self deny: #'xx_omni_x_indexOf:startingAt:' isUnary.
	
	self deny: #// isUnary.
	self deny: #'xx_omni_x_//' isUnary! !

!OstCompilerUtilitiesTest methodsFor: 'testing-symbol' stamp: 'StefanMarr 12/11/2011 16:49'!
testNumArgs
	self assert: #string numArgs = 0.
	self assert: #'xx_omni_x_string' numArgs = 0.
	
	self assert: #linesDo: numArgs = 1.
	self assert: #'xx_omni_x_linesDo:' numArgs = 1.
	
	self assert: #indexOf:startingAt: numArgs = 2.
	self assert: #'xx_omni_x_indexOf:startingAt:' numArgs = 2.
	
	self assert: #// numArgs = 1.
	self assert: #'xx_omni_x_//' numArgs = 1! !

!SequenceableCollection methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/9/2011 17:04'!
omniReplaceFrom: start to: stop with: replacement startingAt: repStart
	| index repOff |
	repOff := repStart - start.
	index := start - 1.
	[ (index := index + 1) <= stop ]
		whileTrue: [ self at: index put: (replacement at: repOff + index) ]! !

!ByteArray methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 18:47'!
replaceFrom: start to: stop with: replacement startingAt: repStart
	"Primitive. This destructively replaces elements from start to stop in the receiver starting at index, repStart, in the collection, replacement. Answer the receiver. Range checks are performed in the primitive only. Optional. See Object documentation whatIsAPrimitive."
	
	<primitive: 105>
	<omniReplacement: #omniReplaceFrom:to:with:startingAt:>
	super replaceFrom: start to: stop with: replacement startingAt: repStart! !

!TraitBehavior methodsFor: '*OmniST-Core-Reflection-override' stamp: 'StefanMarr 12/11/2011 16:49'!
selectorsAndMethodsDo: aBlock
	"Evaluate selectorBlock for all the message selectors in my method dictionary."

	^ self methodDict keysAndValuesDo: [ :selector :method |
		selector isOmniSelector 
			ifFalse: [ aBlock value: selector value: method ] ]! !

Object subclass: #OPromise
	instanceVariableNames: 'signal value'
	classVariableNames: 'NoValueDelivered'
	poolDictionaries: ''
	category: 'OmniST-Core-Misc'!

!OPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/6/2011 00:19'!
example
	"This is not a real test case, but just to see how it is supposed to work"
	| p |
	p := OPromise new.
	
	[|val| 
	 Transcript log: 'About to read promised value.'.
	 val := p read.
	 Transcript log: 'Val: ', val asString.] fork.

	Processor yield.
	
	Transcript log: 'About to deliver promised value.'.
	p deliver: 42.
	Transcript log: 'Deliver promised value.'.
	! !

!OPromise class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/4/2011 15:53'!
initialize
	NoValueDelivered := Object new.! !

!OPromise methodsFor: 'announcing' stamp: 'StefanMarr 11/4/2011 16:07'!
deliver: aValue
	"Deliver the promised value."
	
	
	"Warning: this is racy, but we don't care for the moment.
	 Should be a single produce anyway. Everything else is a bug.
	 And, we do at least best effort error notification..."
	value = NoValueDelivered ifFalse: [
		| err |
		err := Error new.
		err messageText: 'Promise has been delivered more than once.'.
		err signal.
	].

	value := aValue.
	
	signal signal.
	
	^ value.! !

!OPromise methodsFor: 'initialization' stamp: 'StefanMarr 11/4/2011 16:09'!
initialize
	value := NoValueDelivered.
	signal := Semaphore new.! !

!OPromise methodsFor: 'accessing' stamp: 'StefanMarr 11/4/2011 16:01'!
read
	[value = NoValueDelivered] whileTrue: [
		"This is ugly, but I do not see how I can prevent the data
		 race for signals otherwise.
		 In case there are multiple possible consumers everyone
		 but the first will starve when they arrive before the
		 promise was delivered.
		 Now, we just have a delay, but it is guaranteed to return."
		signal waitTimeoutMSecs: 100.
	].
	^ value! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
at: index 
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. See Object documentation 
	whatIsAPrimitive."

	<primitive: 60>
	<omniReplacement: #atomicAt:>
	index isInteger ifTrue:
		[self class isVariable
			ifTrue: [self errorSubscriptBounds: index]
			ifFalse: [self errorNotIndexable]].
	index isNumber
		ifTrue: [^self at: index asInteger]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
at: index put: value
	"Primitive. Assumes receiver is indexable. Store the argument value in 
	the indexable element of the receiver indicated by index. Fail if the 
	index is not an Integer or is out of bounds. Or fail if the value is not of 
	the right type for this kind of collection. Answer the value that was 
	stored. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 61>
	<omniReplacement: #atomicAt:put:>
	index isInteger ifTrue:
		[self class isVariable
			ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self errorImproperStore]
					ifFalse: [self errorSubscriptBounds: index]]
			ifFalse: [self errorNotIndexable]].
	index isNumber
		ifTrue: [^self at: index asInteger put: value]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
basicAt: index
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. Do not override in a subclass. See 
	Object documentation whatIsAPrimitive."

	<primitive: 60>
	<omniReplacement: #atomicBasicAt:>
	index isInteger ifTrue: [self errorSubscriptBounds: index].
	index isNumber
		ifTrue: [^self basicAt: index asInteger]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
basicAt: index put: value
	"Primitive. Assumes receiver is indexable. Store the second argument 
	value in the indexable element of the receiver indicated by index. Fail 
	if the index is not an Integer or is out of bounds. Or fail if the value is 
	not of the right type for this kind of collection. Answer the value that 
	was stored. Essential. Do not override in a subclass. See Object 
	documentation whatIsAPrimitive."

	<primitive: 61>
	<omniReplacement: #atomicBasicAt:put:>
	index isInteger
		ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self errorImproperStore]
					ifFalse: [self errorSubscriptBounds: index]].
	index isNumber
		ifTrue: [^self basicAt: index asInteger put: value]
		ifFalse: [self errorNonIntegerIndex]! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/8/2011 21:32'!
doesNotUnderstand: aMessage 
	 "Handle the fact that there was an attempt to send the given message to the receiver but the receiver does not understand this message (typically sent from the machine when a message is sent to the receiver and no method is defined for that selector)."
	"Testing: (3 activeProcess)"
	"fixed suggested by Eliot miranda to make sure 
	
	[Object new blah + 1]
		on: MessageNotUnderstood
		do: [:e | e resume: 1] does not loop indefinitively"
		
	| exception resumeValue |
	
	<omniDoNotTransform>	
	aMessage selector isOmniSelector
		ifTrue: [
			OstCompiler omniMethod: aMessage missing: self in: aMessage lookupClass.
			
			(self respondsTo: aMessage selector) ifTrue: [
				"And try again."
				^ aMessage sentTo: self.
			]].
		
	(exception := MessageNotUnderstood new)
				message: aMessage;
				receiver: self.
	resumeValue := exception signal.
	^exception reachedDefaultHandler
				ifTrue: [aMessage sentTo: self]
				ifFalse: [resumeValue]! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
halt
	"This is the typical message to use for inserting breakpoints during 
	debugging. It behaves like halt:, but does not call on halt: in order to 
	avoid putting this message on the stack. Halt is especially useful when 
	the breakpoint message is an arbitrary one."

	<omniDoNotTransform>
	Halt signal! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
instVarAt: index 
	"Primitive. Answer a fixed variable in an object. The numbering of the 
	variables corresponds to the named instance variables. Fail if the index 
	is not an Integer or is not the index of a fixed variable. Essential. See 
	Object documentation whatIsAPrimitive."

	<primitive: 73>
	<omniReplacement: #atomicInstVarAt:>
	"Access beyond fixed variables."
	^self basicAt: index - self class instSize		! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
instVarAt: anInteger put: anObject 
	"Primitive. Store a value into a fixed variable in the receiver. The 
	numbering of the variables corresponds to the named instance variables. 
	Fail if the index is not an Integer or is not the index of a fixed variable. 
	Answer the value stored as the result. Using this message violates the 
	principle that each object has sovereign control over the storing of 
	values into its instance variables. Essential. See Object documentation 
	whatIsAPrimitive."

	<primitive: 74>
	<omniReplacement: #atomicInstVarAt:put:>
	"Access beyond fixed fields"
	^self basicAt: anInteger - self class instSize put: anObject! !

!Object methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/9/2011 13:37'!
omniDomain
	^ nil! !

!Object methodsFor: '*OmniST-Core-Reflection' stamp: 'StefanMarr 12/9/2011 15:59'!
perform: aSymbol with: firstObject with: secondObject with: thirdObject with: o4
	"Send the selector, aSymbol, to the receiver with the given arguments.
	Fail if the number of arguments expected by the selector is not three.
	Primitive. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 83>
	^ self perform: aSymbol
		withArguments: (Array with: firstObject with: secondObject with: thirdObject with: o4)! !

!Object methodsFor: '*OmniST-Core-Reflection-overrides' stamp: 'StefanMarr 12/11/2011 16:25'!
shallowCopy
	"Answer a copy of the receiver which shares the receiver's instance variables. It should never be overridden. I'm invoked from the copy template method. Subclasses that need to specialize the copy should specialize the postCopy hook method."
	
	| class newObject index |
	<omniReplacement: #atomicShallowCopy>
	<primitive: 148>
	class := self class.
	class isVariable
		ifTrue: 
			[index := self basicSize.
			newObject := class basicNew: index.
			[index > 0]
				whileTrue: 
					[newObject basicAt: index put: (self basicAt: index).
					index := index - 1]]
		ifFalse: [newObject := class basicNew].
	index := class instSize.
	[index > 0]
		whileTrue: 
			[newObject instVarAt: index put: (self instVarAt: index).
			index := index - 1].
	^ newObject! !

Object subclass: #OstCompilerTestSubject
	instanceVariableNames: 'omniDomain a b c'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler-Test'!

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
a
	^a! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
a: anA
	a := anA! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
aAndB
	^ a, b! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
b
	^b! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
b: anB
	b := anB! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:20'!
initialize
	a := #A.
	b := #B.
	c := #C.! !

!OstCompilerTestSubject methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 16:09'!
omniDomain
	^ omniDomain! !

!OstCompilerTestSubject methodsFor: 'accessing' stamp: 'StefanMarr 11/28/2011 14:21'!
omniDomain: aDomain
	omniDomain := aDomain! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:21'!
return1
	^ 1! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:21'!
returnAViaTemp
	| oldA |
	oldA := a.
	^ oldA! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:21'!
returnNil
	^ nil! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:21'!
returnSelf
	^ self! !

!OstCompilerTestSubject methodsFor: 'methods for testing' stamp: 'StefanMarr 11/28/2011 14:21'!
somethingLongAndComplexReturningOriginalA
	| oldA and to |
	oldA := a.
	
	oldA = a ifTrue: [
		10 timesRepeat: [ a :=	#a ].
		
		and := c.
		to := b asString, c asString.
		
		"Will never happen".
		c == #abba
			ifTrue: [
				[a := self will never happen.].
				c := b + a + 10 factorial.]
			ifFalse: [
				[self this is always going to happen but this block will never be evaluated anyway,
				and we are just poluting the number of literals in this method,
				to make sure that we are not missing anything that should have been done by the compiler
				but is done wrong after we added our changes.].
				
				c := 1 + 2 + 3 + 4 + 5 + 6 + 7 + 8 + 9 + 10.
			]
		].
	
	^ oldA! !

Object subclass: #OstDomain
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Core-Reflection'!

OstDomain subclass: #OstCompilerTestDomain
	instanceVariableNames: 'record'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OmniST-Core-Compiler-Test'!

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:23'!
initialize
	record := OrderedCollection new.! !

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:23'!
readField: idx of: obj
	record add: {#readField:of:. idx. obj}.
	^ obj instVarAt: idx! !

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:31'!
record
	^ record! !

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 16:33'!
requestExecutionOf: selector on: anObj
	record add: { #requestExecutionOf:on:. selector asNormalSelector. anObj }.
	^ anObj perform: selector
! !

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/9/2011 16:33'!
requestExecutionOf: selector on: anObj with: p1
	record add: { #requestExecutionOf:on:with:. selector asNormalSelector. anObj. p1 }.
	^ anObj perform: selector with: p1.
! !

!OstCompilerTestDomain methodsFor: 'as yet unclassified' stamp: 'StefanMarr 11/27/2011 17:24'!
write: val toField: idx of: obj
	record add: {#write:toField:of:. val. idx. obj }.
	obj instVarAt: idx put: val.! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:32'!
classHasDomainField: cls
	^ 0 ~= (self domainIndexFromClass: cls)! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/8/2011 16:36'!
currentDomain
	^ Processor activeProcess omniDomain! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/7/2011 23:08'!
domainIndexFromClass: cls
	^ cls instVarIndexFor: #omniDomain! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:32'!
domainIndexFromObject: obj
	^ self domainIndexFromClass: obj class! !

!OstDomain class methodsFor: 'class initialization' stamp: 'StefanMarr 12/8/2011 16:12'!
initialize
	(Process hasInstVarNamed: #omniDomain) ifFalse: [
		Process addInstVarNamed: #omniDomain.
	].! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:38'!
initializeDomainForNewObject: obj of: class
	(obj respondsTo: #omniDomain:) ifTrue: [
		obj omniDomain: self currentDomain domainForNewObjects.
	].! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:37'!
objectHasDomainField: obj
	^ self classHasDomainField: obj class! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:37'!
omniDomainFrom: obj
	| idx |
	idx := self domainIndexFromObject: obj.
	
	idx ~= 0 ifTrue: [
		^ obj instVarAt: idx.
	].

	^ nil! !

!OstDomain class methodsFor: 'as yet unclassified' stamp: 'StefanMarr 12/11/2011 18:37'!
setOmniDomainOn: obj of: cls to: aDomain
	| idx |
	idx := self domainIndexFromClass: cls.
	obj instVarAt: idx put: aDomain.! !

!OstDomain class methodsFor: 'initialize-release' stamp: 'StefanMarr 12/8/2011 16:32'!
unload
	(Process hasInstVarNamed: #omniDomain) ifTrue: [
		Process removeInstVarNamed: #omniDomain.
	].! !

!OstDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 12/8/2011 14:06'!
adoptObject: obj
	self class setOmniDomainOn: obj of: obj class to: self! !

!OstDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 12/8/2011 14:06'!
domainForNewObjects
	"#basicNew need to determine which domain a new object is supposed
	 to belong to.
	 The standard domain is self, the domain of the currently executing
	 OProcess.
	
	 If this method is overriden, make sure that you do not allocate any objects here!!!!!!"
	^ self! !

!OstDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 12/8/2011 16:37'!
evaluateInContext: aBlock
	| oldDomain proc result |
	proc := Processor activeProcess.
	
	oldDomain := proc omniDomain.
	
	proc omniDomain: self.
	
	result := aBlock value.
	
	proc omniDomain: oldDomain.
	
	^ result! !

!OstDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 12/8/2011 14:06'!
newProcessFor: aBlock
	"Answer a Process running the code in aBlock in this domain.
	 The process is not scheduled."
	| ctx proc |
	<primitive: 19> "Simulation guard"
	ctx := [aBlock value.
			Processor terminateActive] asContext.
	proc := OProcess
		forContext: ctx
		priority: Processor activePriority.
	proc omniDomain: self.
	^ proc.! !

!OstDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/8/2011 14:06'!
readField: idx of: obj
	Transcript cr; show: 'readField: ', idx asString, ' of: ', obj asString.
	^ obj instVarAt: idx! !

!OstDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/9/2011 15:57'!
readGlobal: association
	^ association value! !

!OstDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 12/8/2011 14:07'!
spawnHere: aBlock
	"Spawn the given block as a process in this domain"
	
	self flag: 'WARNING: copied from ODomain.
	TODO: fix later after abstracting concepts from the results of the experiments.'.
	
	^ (self newProcessFor: aBlock) resume! !

!OstDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/8/2011 14:07'!
write: val toField: idx of: obj
	Transcript cr; show: 'write: ', val asString, ' toField: ', idx asString, ' of: ', obj asString.
	^ obj! !

!OstDomain methodsFor: 'omni handlers' stamp: 'StefanMarr 12/9/2011 15:57'!
writeGlobal: association value: val
	^ association value: val! !

!RBLexicalScope methodsFor: '*OmniST-Core-Compiler' stamp: 'StefanMarr 12/9/2011 15:20'!
numOmniTemps
	^ (bindings keys select: [:name | name beginsWith: 'x_omni_']) size! !
OstCompiler initialize!
OPromise initialize!
OstTransformer initialize!
OstDomain initialize!
